#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include "Config.h"
#include "stm32f107.h"
#include "ulcd_defs.h"

static volatile uint32_t __counter;
static volatile uint32_t led_counter;

static inline void delay_ms(uint32_t ms){
  __counter = 0;
  while(__counter < ms * 10);
}

static void init();
void memclr(void *buffer, size_t size);

#define ulcd_delay()
#define ulcd_waitcmd()        while(bit_is_set(ulcd_regread(ULCD_REG_STATUS), 0))

uint32_t ulcd_regread(uint8_t index);
void ulcd_regwrite(uint8_t index, uint32_t value);
void ulcd_regmov(uint8_t to, uint8_t from);
void ulcd_cmd(uint8_t index, uint8_t *param, uint32_t count);

#include "files/timer.h"

int main(){
  init(); // Инициализация переферийных устройств STM32

  delay_ms(500); // Инициализация ULCD
  ulcd_cmd(ULCD_CMD_SOFTRESET, 0, 0); // Программный сброс
  delay_ms(500);

  // Создание контейнера с обработчиком события TIMER.
  ulcd_cmd(ULCD_CMD_BUFFER_CREATE, 0, 0);
  ulcd_waitcmd();
  uint32_t ontimer_buf = ulcd_regread(ULCD_REG_CURRENT);
  // Установка размера контейнера
  ulcd_regwrite(ULCD_REG_OPERAND0, __src_files_timer_scr_len);
  ulcd_cmd(ULCD_CMD_BUFFER_SETSIZE, 0, 0);
  ulcd_waitcmd();
  // Запись данных в контейнер.
  ulcd_regwrite(ULCD_REG_POINTER, 0);
  for(uint32_t i = 0; i < __src_files_timer_scr_len; i++){
    ulcd_regwrite(ULCD_REG_OPERAND0, __src_files_timer_scr[i]);
    ulcd_cmd(ULCD_CMD_BUFFER_PUTBYTE, 0, 0);
    ulcd_waitcmd();
  }

  uint8_t param[13];

  // Создаем таймер
  param[0] = ULCD_WIDGET_TIMER;
  ulcd_cmd(ULCD_CMD_WIDGET_CREATE, param, 1);
  ulcd_waitcmd();
  uint32_t timer = ulcd_regread(ULCD_REG_CURRENT);
  // Устанавливаем позицию
  ulcd_regwrite(ULCD_REG_OPERAND0, 500); // мс
  ulcd_cmd(ULCD_CMD_WIDGET_SETPOSITION, 0, 0);
  ulcd_waitcmd();
  // Установка обработчика события OnTimer.
  param[0] = ULCD_EVT_TIMER;
  ulcd_regwrite(ULCD_REG_OPERAND0, ontimer_buf);
  ulcd_cmd(ULCD_CMD_WIDGET_SETHANDLER, param, 1);
  ulcd_waitcmd();
  // Включаем
  ulcd_regwrite(ULCD_REG_OPERAND0, 1);
  ulcd_cmd(ULCD_CMD_WIDGET_SETENABLED, 0, 0);
  ulcd_waitcmd();

  // Создаем фрейм
  param[0] = ULCD_WIDGET_FRAME;
  ulcd_cmd(ULCD_CMD_WIDGET_CREATE, param, 1);
  ulcd_waitcmd();
  uint32_t frame = ulcd_regread(ULCD_REG_CURRENT);
  ulcd_regwrite(ULCD_REG_R2, frame);

  // Вывод на экран фрейма
  ulcd_regwrite(ULCD_REG_CURRENT, frame);
  ulcd_cmd(ULCD_CMD_SETFRAME, 0, 0);
  ulcd_waitcmd();

  bool ledison = false;
  led_counter = 0;

  uint32_t value = 0;

  while(true){
    // Моргание светодиодом
    if(led_counter > 500){
      if(ledison){
        GPIOC_BSRR = _BV(9);
      }
      else {
        GPIOC_BRR = _BV(9);
      }
      ledison = !ledison;
      led_counter = 0;
    }
    // Регистр увеличивается на 1 каждые 500мс
    value = ulcd_regread(ULCD_REG_R3);
  }
}

// Прерывание системного таймера (SysTick)
void __SysTick(){
  __counter++;
  led_counter++;
}

uint8_t spi_com(uint8_t data){
  GPIOD_BRR = _BV(7);
  while(bit_is_clear(SPI3_SR, SPI3_SR_TXE));
  SPI3_DR = data;
  while(bit_is_clear(SPI3_SR, SPI3_SR_RXNE));
  data = SPI3_DR;
  GPIOD_BSRR = _BV(7);
  return data;
}

uint32_t ulcd_regread(uint8_t index){
  spi_com(ULCD_CMD_REG_READ);
  ulcd_delay();
  spi_com(index);
  ulcd_delay();
  delay_ms(1);
  uint32_t result = 0;
  for(uint32_t i = 0; i < 4; i++){
    uint32_t data = (uint32_t)spi_com(0x00);
    result |= (data << (8 * i));
    ulcd_delay();
  }
  return result;
}

void ulcd_regwrite(uint8_t index, uint32_t value){
  spi_com(ULCD_CMD_REG_WRITE);
  ulcd_delay();
  spi_com(index);
  ulcd_delay();
  for(uint32_t i = 0; i < 4; i++){
    spi_com((value >> (8 * i)) & 0xFF);
    ulcd_delay();
  }
}

void ulcd_regmov(uint8_t to, uint8_t from){
  spi_com(ULCD_CMD_REG_MOVE);
  ulcd_delay();
  spi_com(to);
  ulcd_delay();
  spi_com(from);
  ulcd_delay();
}

void ulcd_cmd(uint8_t index, uint8_t *param, uint32_t count){
  spi_com(index);
  ulcd_delay();
  for(uint32_t i = 0; i < count; i++){
    spi_com(param[i]);
    ulcd_delay();
  }
}

// Инициализация переферийных устройств
static void init(){
  // Инициализация системного таймера
  SysTickReload = TIMRELOAD_CALC(100, 72000000); // 100 мкс ~ 10кГц
  SysTickCurrent = 0;
  SysTickControl = _BV(SysTickControl_ENABLE) | _BV(SysTickControl_TICKINT) |
  _BV(SysTickControl_CLKSOURCE); // Источник тактирования -- частота
  // процессора (AHB)

  RCC_APB2ENR |= _BV(RCC_APB2ENR_AFIOEN) | _BV(RCC_APB2ENR_IOPAEN) |
  _BV(RCC_APB2ENR_IOPBEN) | _BV(RCC_APB2ENR_IOPCEN) | _BV(RCC_APB2ENR_IOPDEN);
  
  // ULCD запрос на прерывание
  GPIOB_CRL = (GPIOB_CRL & ~((3 << GPIOB_CRL_CNF7) | (3 << GPIOB_CRL_MODE7))) |
  ((1 << GPIOB_CRL_CNF7) | (0 << GPIOB_CRL_MODE7)); // PB7 -- плавающий вход

  // Конфигурация выводов SPI и ULCD

  GPIOB_CRL = (GPIOB_CRL & ~((3 << GPIOB_CRL_CNF7) | (3 << GPIOB_CRL_MODE7))) |
  ((1 << GPIOB_CRL_CNF7) | (0 << GPIOB_CRL_MODE7)); // PB7, INT -- Плавающий
  // вход

  GPIOD_CRL = (GPIOD_CRL & ~((3 << GPIOD_CRL_CNF7) | (3 << GPIOD_CRL_MODE7))) |
  ((0 << GPIOD_CRL_CNF7) | (3 << GPIOD_CRL_MODE7)); // PD7, SSEL --
  // Двухтактный выход, 50МГц

  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF10) |
  (3 << GPIOC_CRH_MODE10))) | ((2 << GPIOC_CRH_CNF10) |
  (3 << GPIOC_CRH_MODE10)); // PC10, SCK -- Альтернативный двухтактный выход

  GPIOC_BSRR = _BV(11);
  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF11) |
  (3 << GPIOC_CRH_MODE11))) | ((2 << GPIOC_CRH_CNF11) |
  (0 << GPIOC_CRH_MODE11)); // PC11, MISO -- Подтянутый к питанию вход

  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF12) |
  (3 << GPIOC_CRH_MODE12))) | ((2 << GPIOC_CRH_CNF12) |
  (3 << GPIOC_CRH_MODE12)); // PC12, MOSI -- Альтернативный двухтактный выход

  AFIO_MAPR |= (1 << AFIO_MAPR_SPI3_REMAP);

  // Инициализация SPI
  RCC_APB1ENR |= _BV(RCC_APB1ENR_SPI3EN);
  SPI3_CR1 = (1 << SPI3_CR1_CPHA) | (1 << SPI3_CR1_CPOL) |
  (1 << SPI3_CR1_MSTR) | SPI3_CR1_BR_PCLK256_gc | (1 << SPI3_CR1_SPE) |
  (0 << SPI3_CR1_LSBFIRST) | (0 << SPI3_CR1_DFF) | (1 << SPI3_CR1_SSM) |
  (1 << SPI3_CR1_SSI);

  GPIOD_BSRR = _BV(7); // SSEL = 1

  // Инициализация светодиода (PC9)
  GPIOC_CRH = (GPIOC_CRH & ~(0xF << 4)) | (0x1 << 4); // Двухтактный
  // выход, 10МГц
  GPIOC_BRR = _BV(9); // PC9 = 0
}

void memclr(void *buffer, size_t size){
  for(uint32_t i = 0; i < size; i++){
    *((uint8_t *)buffer++) = 0;
  }
}

